from pydantic_typer import Typer

from abertpy.ping import app as ping_app
from abertpy.proxy import app as proxy_app
from abertpy.setup import app as setup_app

app = Typer()


app.add_typer(setup_app)
app.add_typer(proxy_app)
app.add_typer(ping_app)



import aiohttp


async def tvh_get_networks(session: aiohttp.ClientSession, base_url: str):
    networks_url = base_url + "/api/mpegts/network/grid"
    async with session.post(
        networks_url,
        data={
            "limit": 99999,
        },
    ) as response:
        networks: dict = await response.json()
        return networks


async def tvh_get_muxes(session: aiohttp.ClientSession, base_url: str):
    networks_url = base_url + "/api/mpegts/mux/grid"
    async with session.post(
        networks_url,
        data={
            "limit": 99999,
        },
    ) as response:
        muxes: dict = await response.json()
        return muxes



import sys
from venv import logger

import backoff
import requests
from pydantic_typer import Typer

from abertpy.models import ProxyArgs

app = Typer()

######################################
######################################
######################################
######################################


FRAME_SIZE = 188
MPEG_TS_START_BYTE = 0x47

AFC_PAYLOAD_ONLY = 0x10
AFC_ADAPTATION_PAYLOAD = 0x30


def process_data(packet: bytes, allowed_pid: int):
    if not packet or packet[0] != MPEG_TS_START_BYTE:
        logger.error("MPEG-TS first byte unexpected")
        return

    try:
        afc = packet[3] & 0x30
    except IndexError as e:
        logger.error(e)
        return

    tspid = ((packet[1] & 0x1F) << 8) | packet[2]
    if tspid != allowed_pid:
        return

    payload: bytes | None = None

    if afc == AFC_PAYLOAD_ONLY:
        payload = bytes(packet[4:])
    elif afc == AFC_ADAPTATION_PAYLOAD:
        payload = bytes(
            packet[packet[4] + 5 :]
        )  # 4 bytes header + 1 byte length indicator
    else:
        logger.error("AFC bad value {}", afc)
        return

    if payload:
        sys.stdout.buffer.write(payload)


@backoff.on_exception(backoff.expo, ConnectionError, max_time=10)
@app.command(help="Proxy for Abertis streams")
def proxy(arg: ProxyArgs):

    base_url = arg.get_base_url()
    endpoint = f"{base_url}/stream/service/{arg.service_uuid}"

    response = requests.get(
        endpoint, stream=True, headers={"User-Agent": "curl/aiohttp"}
    )
    for packet in response.iter_content(chunk_size=FRAME_SIZE):
        process_data(packet=packet, allowed_pid=arg.allowed_pid)



import asyncio
import json
import subprocess

import aiohttp
from loguru import logger
from pydantic_typer import Typer

from abertpy.helpers import tvh_get_muxes, tvh_get_networks
from abertpy.models import SetupArgs

app = Typer()

_HARDCODED_KEY = "abertpy"
_HARDCODED_PMT = 8000


_MAP_PPID_CA: dict[int, int] = {}


async def get_muxes(session: aiohttp.ClientSession, arg: SetupArgs) -> list[dict]:

    muxes = await tvh_get_muxes(session, arg.get_base_url())

    target_muxes: list[dict] = [
        mux
        for mux in muxes.get("entries", [])
        if mux.get("enabled", True)
        and mux.get("network_uuid", None) == arg.network_uuid
    ]

    logger.info(f"Target muxes: {len(target_muxes)}")
    return target_muxes


async def fetch_mux_data(
    session: aiohttp.ClientSession,
    url: str,
    arg: SetupArgs,
    buffer: bytearray,
):
    async with session.get(url) as response:
        total_bytes = 0
        async for data in response.content.iter_chunked(1024 * 10):
            total_bytes += len(data)
            buffer.extend(data)
            if total_bytes >= arg.mux_buffer_size:
                logger.debug(f"Stopped after receiving {total_bytes} bytes.")
                return


async def get_mux_data(
    session: aiohttp.ClientSession,
    arg: SetupArgs,
    url: str,
):
    buffer: bytearray = bytearray()
    try:
        await asyncio.wait_for(
            fetch_mux_data(session, url, arg, buffer),
            timeout=arg.mux_buffer_time.total_seconds(),
        )
    except asyncio.TimeoutError:
        pass

    process = subprocess.Popen(
        [str(arg.tsanalyze_path), "--json"],
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    stdout, _ = process.communicate(input=buffer)
    tsanlyze_output = json.loads(stdout)

    return tsanlyze_output


async def create_iptv_network(session: aiohttp.ClientSession, arg: SetupArgs) -> str:

    # Find network with pnetworkname containing abertpy
    networks = await tvh_get_networks(session, arg.get_base_url())

    for network in networks.get("entries", []):
        if _HARDCODED_KEY in network.get("networkname", ""):
            logger.info(f"IPTV Network already exist: {network.get('networkname')}")
            return network.get("uuid")

    async with session.post(
        arg.get_base_url() + "/api/mpegts/network/create",
        data={
            "class": "iptv_network",
            "conf": json.dumps(
                {
                    "enabled": True,
                    "networkname": f"{_HARDCODED_KEY}: Abertis",  # Keyword to find it later
                    "bouquet": False,
                    "max_streams": 0,
                    "max_bandwidth": 0,
                    "pnetworkname": "",
                    "nid": 0,
                    "ignore_chnum": False,
                    "satip_source": 0,
                    "charset": "",
                    "use_libav": False,
                    "scan_create": True,
                    "priority": 1,
                    "spriority": 1,
                    "max_timeout": 15,
                    "icon_url": "",
                    "skipinitscan": True,
                    "idlescan": False,
                    "sid_chnum": False,
                    "localtime": 0,
                    "service_sid": 0,
                    "remove_scrambled": True,
                }
            ),
        },
    ) as response:
        net_uuid: dict[str, str] = await response.json()
        return net_uuid["uuid"]


async def tvh_find_service_uuid(
    session: aiohttp.ClientSession,
    arg: SetupArgs,
    mux_uuid: str,
    service_sid: int,
) -> str:
    async with session.post(
        arg.get_base_url() + "/api/mpegts/service/grid",
        data={
            "hidemode": "none",
            "filter": json.dumps(
                [
                    {
                        "type": "numeric",
                        "comparison": "eq",
                        "field": "sid",
                        "value": service_sid,
                    }
                ]
            ),
        },
    ) as response:
        resp = await response.json()

    for service in resp.get("entries", []):
        if service.get("multiplex_uuid", "") == mux_uuid:
            return service["uuid"]

    raise ValueError(f"Service UUID not found for SID {service_sid} in mux {mux_uuid}")


async def tvh_override_exists(
    session: aiohttp.ClientSession,
    arg: SetupArgs,
    mux_uuid: str,
    private_pid: int,
) -> str | None:
    async with session.post(
        arg.get_base_url() + "/api/mpegts/service/grid",
        data={
            "hidemode": "none",
            "filter": json.dumps(
                [
                    {
                        "type": "string",
                        "field": "svcname",
                        "value": _HARDCODED_KEY,
                    }
                ]
            ),
        },
    ) as response:
        resp = await response.json()

    for service in resp.get("entries", []):
        if (
            service.get("multiplex_uuid", "") == mux_uuid
            and service.get("sid", -1) == private_pid
            and _HARDCODED_KEY in service.get("svcname", "")
        ):
            return service["uuid"]


async def recreate_tvh_service(
    session: aiohttp.ClientSession,
    arg: SetupArgs,
    mux_uuid: str,
    private_pid: int,
    service_sid: int,
) -> str:

    svc_uuid: str | None = None

    # Extract original SID always.
    tvh_original_uuid: str = await tvh_find_service_uuid(
        session, arg, mux_uuid, service_sid
    )
    tvh_overriden_uuid: str | None = await tvh_override_exists(
        session, arg, mux_uuid, private_pid
    )

    svc_uuid = tvh_overriden_uuid or tvh_original_uuid

    async with session.get(
        arg.get_base_url() + "/api/raw/export",
        params={
            "uuid": svc_uuid,
        },
    ) as response:
        res = await response.json()

    sid_original = (res)[0]

    if pcr := sid_original.get("pcr", None):
        _MAP_PPID_CA[private_pid] = pcr

    # Avoid duplicates
    if tvh_overriden_uuid:
        return tvh_overriden_uuid

    logger.debug(f"Original SID data: {sid_original}")

    sid_original["stream"].append(
        {
            "pid": 8191,
            "type": "CA",
            "position": 0,
            "caidlist": [{"caid": 9728}, {"caid": 9728}],
        }
    )

    # Hijack Tvheadend
    sid_original["sid"] = private_pid
    sid_original["verified"] = (
        1  # Very important ortherwise tvheadend will not stream data
    )
    sid_original["pmt"] = _HARDCODED_PMT  # Always 8000?
    sid_original["stream"].append({"type": "H264", "position": 0, "pid": private_pid})
    sid_original["svcname"] = (
        f"{_HARDCODED_KEY}: raw pPID {private_pid} (SID: {service_sid})"  # pd stands for private data
    )
    sid_original["enabled"] = True

    async with session.post(
        arg.get_base_url() + "/api/raw/import",
        data={
            "node": json.dumps(sid_original),
        },
    ) as response:
        pass

    return tvh_original_uuid


async def recreate_tvh_iptv_mux(
    session: aiohttp.ClientSession,
    arg: SetupArgs,
    iptv_network_uuid: str,
    svc_mux_uuid: str,
    private_pid: int,
    mux_freq: str,
):

    muxes = await tvh_get_muxes(session, arg.get_base_url())

    for mux in muxes.get("entries", []):
        if (
            mux.get("network_uuid", None) == iptv_network_uuid
            and _HARDCODED_KEY in mux.get("iptv_muxname", "")
            and f"{private_pid}" in mux.get("iptv_muxname", "")
        ):
            return

    async with session.post(
        arg.get_base_url() + "/api/mpegts/network/mux_create",
        data={
            "uuid": iptv_network_uuid,
            "conf": json.dumps(
                {
                    "enabled": 1,
                    "epg": 1,
                    "epg_module_id": "",
                    "iptv_url": arg.get_iptv_pipe(
                        svc_mux_uuid=svc_mux_uuid, allowed_pid=private_pid
                    ),
                    "use_libav": 0,
                    "iptv_atsc": False,
                    "iptv_muxname": f"{_HARDCODED_KEY}: MUX {mux_freq} pPID {private_pid}",
                    "channel_number": "0",
                    "iptv_sname": "",
                }
            ),
        },
    ) as response:
        pass


######################################
######################################
######################################
######################################


async def setup_async(arg: SetupArgs):
    base_url = arg.get_base_url()

    async with aiohttp.ClientSession(
        raise_for_status=True,
        headers={
            "User-Agent": "curl/aiohttp"
        },  # https://docs.tvheadend.org/documentation/development/json-api/other-functions#play
    ) as session:

        # First thing, create a IPTV Network if not existing
        abertis_net_uuid = await create_iptv_network(session, arg)

        # Get enabled muxes from tvheadend
        list_muxes = await get_muxes(session, arg)

        map_dataPID_SID: dict[int, int] = {}

        for mux in list_muxes:
            mux_uuid = mux["uuid"]
            mux_freq: str = mux.get("name", "")

            logger.debug(f"Scanning mux: {mux_uuid} - {mux_freq}")
            tsanalyzer_dict = await get_mux_data(
                session,
                arg,
                f"{base_url}/play/ticket/stream/mux/{mux.get('uuid')}",
            )

            found_p_pid = []

            for pid in tsanalyzer_dict.get("pids", []):
                # Skip PMT
                if pid["pmt"]:
                    continue

                # Skip FTA
                if not pid["is-scrambled"]:
                    continue

                # Skip audio/video
                if pid["audio"] or pid["video"]:
                    continue

                # No lang
                if pid.get("language", None):
                    continue

                # Must have sercvices. TODO: maybe more than 1 svc?
                if pid["service-count"] != 1:
                    continue

                logger.debug(f"PID: {pid}")

                # Associate Private data PID to SID
                abertis_data_pid = pid["id"]
                service_sid = pid["services"][0]

                found_p_pid.append(abertis_data_pid)

                svc_mux_uuid = await recreate_tvh_service(
                    session,
                    arg,
                    mux_uuid,
                    private_pid=abertis_data_pid,
                    service_sid=service_sid,
                )

                await recreate_tvh_iptv_mux(
                    session,
                    arg,
                    iptv_network_uuid=abertis_net_uuid,
                    svc_mux_uuid=svc_mux_uuid,
                    private_pid=abertis_data_pid,
                    mux_freq=mux_freq,
                )

                map_dataPID_SID[abertis_data_pid] = service_sid

            logger.info(
                "MUX {} private pids: {}",
                mux_freq,
                ",".join(str(_) for _ in sorted(found_p_pid)),
            )

    for p_pid, pid_ca in sorted(_MAP_PPID_CA.items()):
        logger.info(
            f"F {p_pid:04X}{pid_ca:04X} 00000000 FFFFFFFFFFFFFFFF ;ABERTIS-abertpy {p_pid} (30.0W)"
        )


@app.command(help="Install everything on TVHeadend")
def setup(arg: SetupArgs):
    logger.info("Setup arguments:\n{}", arg.model_dump_json(indent=2))

    return asyncio.run(setup_async(arg))



import asyncio
import shutil
import subprocess
import sys
from datetime import timedelta
from pathlib import Path
from typing import Annotated, Self

import aiohttp
import pydantic
import typer
from loguru import logger
from pydantic import ByteSize, Field

from abertpy.helpers import tvh_get_networks

_REFERENCE_PING = "ping"
_REFERENCE_PROXY = "proxy"


class CommonArgs(pydantic.BaseModel):
    model_config = pydantic.ConfigDict(validate_default=True)

    debug: Annotated[
        bool,
        typer.Option(
            "-d",
            "--debug",
            help="Debug info",
        ),
    ] = False

    tvheadend_url: Annotated[
        pydantic.HttpUrl,
        typer.Option(
            "-t",
            "--tvhurl",
            help="Base URL of the TVheadend server. Ex: http://tvheadend.lan:9981/doesnt_matter_the_path",
            metavar="URL",
        ),
    ]

    @pydantic.field_validator("debug")
    @classmethod
    def set_debug(cls, debug):
        logger.remove()

        logger.add(sys.stderr, level="DEBUG" if debug else "INFO")

        return debug

    @pydantic.field_validator("tvheadend_url")
    @classmethod
    def validate_url(cls, tvheadend_url):
        async def validate_tvheadend_url(tvheadend_url):
            base_url = str(tvheadend_url).removesuffix(tvheadend_url.path or "/")
            serverinfo_url = base_url + "/api/mpegts/mux/grid"

            async with aiohttp.ClientSession() as session:
                async with session.get(serverinfo_url) as response:
                    if response.status != 200:
                        raise ValueError(
                            f"TVheadend error {response.status}. Check user credentials and access permissions."
                        )
            return tvheadend_url

        return asyncio.run(validate_tvheadend_url(tvheadend_url))

    def get_base_url(self) -> str:
        return str(self.tvheadend_url).removesuffix(self.tvheadend_url.path or "/")


class ProxyArgs(CommonArgs):
    model_config = pydantic.ConfigDict(validate_default=True)

    service_uuid: Annotated[
        str,
        typer.Option(
            "-s",
            "--service",
            help="UUID of the Abertis service to proxy",
        ),
    ]

    allowed_pid: Annotated[
        int,
        typer.Option(
            "-a",
            "--allowed-pids",
            help="Allowed MPEG TS PID for decapsulation",
        ),
    ]

    @pydantic.model_validator(mode="after")
    def validate_service_uuid(self):
        async def validate_tvheadend_url(tvheadend_url, service_uuid):
            base_url = str(tvheadend_url).removesuffix(tvheadend_url.path or "/")
            serverinfo_url = base_url + "/api/mpegts/service/grid"

            async with aiohttp.ClientSession() as session:
                async with session.post(
                    serverinfo_url,
                    data={"hidemode": "none", "limit": 10000},
                ) as response:
                    if response.status != 200:
                        raise ValueError(
                            f"TVheadend error {response.status}. Check user credentials and access permissions."
                        )
                    resp = await response.json()

            svcs: list = resp.get("entries", [])
            for service in svcs:
                if service.get("uuid", "") == service_uuid:
                    return service_uuid

            raise ValueError(
                f"Service UUID {service_uuid} not found in TVheadend ({len(svcs)} services scanned)."
            )

        asyncio.run(validate_tvheadend_url(self.tvheadend_url, self.service_uuid))
        return self


class SetupArgs(CommonArgs):
    model_config = pydantic.ConfigDict(validate_default=True)

    network_uuid: Annotated[
        str | None,
        typer.Option(
            "-n",
            "--network-uuid",
            help="DVB-S Network containing Abertis muxes (usually Hispasat 30W). Empty to list all networks.",
        ),
    ] = None

    tsanalyze_path: Annotated[
        Path | None,
        typer.Option(
            " ",
            "--path-tsanalyze",
            help="Path to the tsanalyze binary from TSDuck. Will search for 'tsanalyze' by default",
            exists=True,
            file_okay=True,
            dir_okay=False,
            writable=False,
            readable=True,
            resolve_path=True,
        ),
    ] = None

    mux_buffer_size: Annotated[
        ByteSize,
        typer.Option(
            " ",
            "--max-buffer-size",
            help="Amount of data to buffer from each mux before analyzing",
            metavar="SIZE",
        ),
    ] = Field(
        default="50MB"
    )  # type: ignore

    mux_buffer_time: Annotated[
        timedelta,
        typer.Option(
            " ",
            "--max-buffer-time",
            help="Maximum time to wait for buffering each mux before analyzing",
            metavar="ISO 8601",
        ),
    ] = "PT10S"  # type: ignore

    abertpy_path: Annotated[
        Path | None,
        typer.Option(
            " ",
            "--path-abertpy",
            help="Command to execute for self referencing abertpyin IPTV mux. It will be injected later for TVHeadend IPTV proxy",
        ),
    ] = Path("/usr/local/bin/abertpy")

    abertpy_validate_binary: Annotated[
        bool,
        typer.Option(
            " ",
            "--validate-abertpy/--no--validate-abertpy",
            help="Validate the abertpy binary. Disable if running outside the TVHeadend host",
        ),
    ] = True

    iptv_pipe_string: Annotated[
        str,
        typer.Option(
            " ",
            "--pipe-command",
            help="""DVB-S Network containing Abertis muxes (usually Hispasat 30W). Empty to list all networks. Allowed variables:
            \n
            * abertpy_path: Full path to abertpy command\n
            * allowed_pid: Private PID number of the REMUX\n
            * svc_mux_uuid: UUID of the service REMUX\n
            * tvheadend_url: Path for tvheadend_url base URL
            """,
        ),
    ] = "pipe://{abertpy_path} proxy -a {allowed_pid} -t http://127.0.0.1:9981/ -s {svc_mux_uuid}"

    @pydantic.model_validator(mode="after")
    def validate_abertpy_path(self):
        if self.abertpy_validate_binary:

            try:
                full_cmd: list[str] = [str(self.abertpy_path)]
                full_cmd.append(_REFERENCE_PING)
                ret = subprocess.run(full_cmd, capture_output=True, text=True)
                if "vk496" not in ret.stdout or ret.returncode != 18:
                    raise ValueError(f"Bad abertpy reference:\n {ret}")

            except Exception as e:
                raise ValueError(
                    f"Bad abertpy binary. Consider disabling this validation"
                ) from e

        return self

    @pydantic.field_validator("tsanalyze_path")
    @classmethod
    def validate_tsduck(cls, tsanalyze_path):
        bin_path: str | None = (
            str(tsanalyze_path) if tsanalyze_path else shutil.which("tsanalyze")
        )

        if not bin_path:
            raise ValueError(
                "tsanalyze binary not found. Please install TSDuck (https://tsduck.io/) or provide the argument path to tsanalyze"
            )

        try:
            ret = subprocess.run(
                [bin_path, "--version"], capture_output=True, text=True
            )
            if "TSDuck" not in ret.stderr:
                raise ValueError(f"Wrong tsanalyze binary?. {ret}")
        except Exception as e:
            raise ValueError(f"Error running tsanalyze") from e

        return Path(bin_path)

    @pydantic.model_validator(mode="after")
    def validate_network_uuid(self) -> Self:
        async def validate_network():
            base_url = self.get_base_url()

            async with aiohttp.ClientSession() as session:
                networks = await tvh_get_networks(session, base_url)

                all_networks: dict[str, str] = {
                    network["uuid"]: network["networkname"]
                    for network in networks.get("entries", [])
                }

                if self.network_uuid is None or self.network_uuid not in all_networks:
                    raise ValueError(
                        "Missing network argument. Available UUID networks:\n"
                        + "\n".join(
                            f"{net_uuid} ({net_name})"
                            for net_uuid, net_name in all_networks.items()
                        )
                        + "\n"
                    )

        asyncio.run(validate_network())
        return self

    def get_iptv_pipe(self, svc_mux_uuid: str, allowed_pid: int) -> str:
        return self.iptv_pipe_string.format(
            abertpy_path=self.abertpy_path,
            tvheadend_url=self.tvheadend_url,
            svc_mux_uuid=svc_mux_uuid,
            allowed_pid=allowed_pid,
        )
